/*-
 *  Copyright 2020 Beuth Hochschule für Technik Berlin, Hochschule für Technik Stuttgart
 * 
 *  This file is part of CityDoctor2.
 *
 *  CityDoctor2 is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  CityDoctor2 is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with CityDoctor2.  If not, see <https://www.gnu.org/licenses/>.
 */
package de.hft.stuttgart.citydoctor2.healing;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertFalse;
import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.assertTrue;
import static org.mockito.Mockito.mock;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;

import org.citygml4j.core.model.core.AbstractThematicSurface;
import org.citygml4j.core.model.core.CityModel;
import org.junit.Rule;
import org.junit.Test;
import org.junit.rules.TemporaryFolder;

import de.hft.stuttgart.citydoctor2.check.CheckError;
import de.hft.stuttgart.citydoctor2.check.Checker;
import de.hft.stuttgart.citydoctor2.check.ModificationListener;
import de.hft.stuttgart.citydoctor2.check.Requirement;
import de.hft.stuttgart.citydoctor2.check.RequirementConfiguration;
import de.hft.stuttgart.citydoctor2.check.ValidationConfiguration;
import de.hft.stuttgart.citydoctor2.check.error.AttributeMissingError;
import de.hft.stuttgart.citydoctor2.datastructure.BoundarySurface;
import de.hft.stuttgart.citydoctor2.datastructure.Building;
import de.hft.stuttgart.citydoctor2.datastructure.CityDoctorModel;
import de.hft.stuttgart.citydoctor2.datastructure.ConcretePolygon;
import de.hft.stuttgart.citydoctor2.datastructure.Geometry;
import de.hft.stuttgart.citydoctor2.datastructure.GeometryType;
import de.hft.stuttgart.citydoctor2.datastructure.Installation;
import de.hft.stuttgart.citydoctor2.datastructure.LinearRing;
import de.hft.stuttgart.citydoctor2.datastructure.LinearRing.LinearRingType;
import de.hft.stuttgart.citydoctor2.datastructure.LinkedPolygon;
import de.hft.stuttgart.citydoctor2.datastructure.Lod;
import de.hft.stuttgart.citydoctor2.datastructure.Polygon;
import de.hft.stuttgart.citydoctor2.datastructure.Vertex;
import de.hft.stuttgart.citydoctor2.exceptions.CityDoctorWriteException;
import de.hft.stuttgart.citydoctor2.healer.Healer;
import de.hft.stuttgart.citydoctor2.parser.CityGmlParseException;
import de.hft.stuttgart.citydoctor2.parser.CityGmlParser;
import de.hft.stuttgart.citydoctor2.parser.InvalidGmlFileException;

public class HealMissingSolidTest {

	@Rule
	public TemporaryFolder folder = new TemporaryFolder();

	@Test
	public void testCreateSolid() throws CityDoctorWriteException, IOException, CityGmlParseException, InvalidGmlFileException {
		ValidationConfiguration config = ValidationConfiguration.loadStandardValidationConfig();
		CityDoctorModel model = new CityDoctorModel(config.getParserConfiguration(), new File("test"));
		model.setCityModel(new CityModel());
		Building b = new Building();
		model.addBuilding(b);
		b.setGmlObject(new org.citygml4j.core.model.building.Building());

		Geometry geom = new Geometry(GeometryType.MULTI_SURFACE, Lod.LOD2);
		b.addGeometry(geom);
		ConcretePolygon p1 = new ConcretePolygon();
		LinearRing ext1 = createTestRing();
		p1.setExteriorRing(ext1);
		geom.addPolygon(p1);

		BoundarySurface bs1 = new BoundarySurface(mock(AbstractThematicSurface.class));
		b.addBoundarySurface(bs1);
		Geometry geom2 = new Geometry(GeometryType.MULTI_SURFACE, Lod.LOD2);
		bs1.addGeometry(geom2);
		ConcretePolygon p2 = new ConcretePolygon();
		LinearRing ext2 = createTestRing();
		p2.setExteriorRing(ext2);
		geom2.addPolygon(p2);
		LinkedPolygon p5 = new LinkedPolygon(p2, geom);
		geom.addPolygon(p5);

		Installation bi = new Installation();
		b.addBuildingInstallation(bi);
		Geometry geom3 = new Geometry(GeometryType.MULTI_SURFACE, Lod.LOD2);
		bi.addGeometry(geom3);
		ConcretePolygon p3 = new ConcretePolygon();
		LinearRing ext3 = createTestRing();
		p3.setExteriorRing(ext3);
		geom3.addPolygon(p3);

		BoundarySurface bs2 = new BoundarySurface(mock(AbstractThematicSurface.class));
		bi.addBoundarySurface(bs2);
		Geometry geom4 = new Geometry(GeometryType.MULTI_SURFACE, Lod.LOD2);
		bs2.addGeometry(geom4);
		ConcretePolygon p4 = new ConcretePolygon();
		LinearRing ext4 = createTestRing();
		p4.setExteriorRing(ext4);
		geom4.addPolygon(p4);

		AttributeMissingError err = new AttributeMissingError(b, "", "lod2solid");

		HealMissingSolid lodHealer = new HealMissingSolid();
		lodHealer.visit(err, mock(ModificationListener.class));

		assertEquals(1, b.getGeometries().size());
		Geometry geometry = b.getGeometries().get(0);
		assertEquals(4, geometry.getPolygons().size());
		assertEquals(Lod.LOD2, geometry.getLod());
		assertEquals(GeometryType.SOLID, geometry.getType());
		int nrOfConcrete = 0;
		int nrOfLinks = 0;
		for (Polygon p : geometry.getPolygons()) {
			if (p.isLink()) {
				nrOfLinks++;
			} else {
				nrOfConcrete++;
			}
		}
		assertEquals(1, nrOfConcrete);
		assertEquals(3, nrOfLinks);

		File createSolidFile = folder.newFile("createdSolid.gml");
		model.saveAs(createSolidFile.toString(), false);
		
		CityDoctorModel parsedModel = CityGmlParser.parseCityGmlFile(createSolidFile.toString(), config.getParserConfiguration());
		Building building = parsedModel.getBuildings().get(0);
		assertEquals(1, building.getGeometries().size());
		Geometry geometry2 = building.getGeometries().get(0);
		assertEquals(GeometryType.SOLID, geometry2.getType());
		
	}
	
	private LinearRing createTestRing() {
		LinearRing ext = new LinearRing(LinearRingType.EXTERIOR);
		ext.addVertex(new Vertex(1, 0, 0));
		ext.addVertex(new Vertex(2, 2, 0));
		ext.addVertex(new Vertex(3, 4, 0));
		ext.addVertex(new Vertex(1, 0, 0));
		return ext;
	}

	@Test
	public void testFixingFile() throws CityGmlParseException, InvalidGmlFileException {
		ValidationConfiguration config = ValidationConfiguration.loadStandardValidationConfig();
		config.setSchematronFilePathInGlobalParameters("src/test/resources/schematronTestForMissingSolid.xml");
		config.getRequirements().put(Requirement.R_SE_BS_ROOF_UNFRAGMENTED.getId(),
				new RequirementConfiguration(false));
		CityDoctorModel model = CityGmlParser.parseCityGmlFile("src/test/resources/SimpleSolid_SrefBS_withoutSolid.gml",
				config.getParserConfiguration());
		Building b = model.getBuildings().get(0);
		assertTrue(b.getGeometries().isEmpty());
		Checker c = new Checker(config, model);
		c.runChecks();
		assertTrue(b.containsAnyError());
		List<CheckError> errors = new ArrayList<>();
		b.collectContainedErrors(errors);
		assertEquals(1, errors.size());
		CheckError err = errors.get(0);
		assertTrue(err instanceof AttributeMissingError);
		Healer healer = new Healer(c);
		healer.healModel(model);
		assertFalse(b.containsAnyError());
		Geometry geom = b.getGeometry(GeometryType.SOLID, Lod.LOD2);
		assertNotNull(geom);
		assertEquals(7, geom.getPolygons().size());
	}

}
